/* $Id: papi_atkdocumentiface.c,v 1.10 2006/07/17 16:24:20 marcusva Exp $
 *
 * Copyright (c) 2006, Marcus von Appen
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "papi.h"
#include "papi_private.h"

/* Forward declarations. */
static PyObject* _not_implemented_noargs (PyObject *self);
static const gchar* _document_get_document_type (AtkDocument *document);
static gpointer _document_get_document (AtkDocument *document);

/**
 * Interface methods, which are bound to the object at runtime.
 */
static PyMethodDef _atkdocumentiface_methods[] =
{
    { "document_get_document_type", (PyCFunction) _not_implemented_noargs,
      METH_NOARGS,
      "O.document_get_document_type () -> string\n\n"
      "Gets a string indicating the document type.\n\n"
      "This method has to be implemented by inherited classes." },
    { "document_get_document", (PyCFunction) _not_implemented_noargs,
      METH_NOARGS,
      "O.document_get_document () -> object\n\n"
      "Gets a reference that points to an instance of the DOM.\n\n"
      "It is up to the caller to check get_document_type to determine how to "
      "cast this reference.\n\n"
      "This method has to be implemented by inherited classes." },
    { NULL, NULL, 0, NULL }
};

/**
 * Default implementation for the interfaces.
 */
static PyObject*
_not_implemented_noargs (PyObject *self)
{
    PyErr_SetString (PyExc_NotImplementedError, "method not implemented");
    return NULL;
}

/**
 * Wrapper for atk_document_get_document_type().
 */
static const gchar*
_document_get_document_type (AtkDocument *document)
{
    PyObject *result;
    PyObject *obj;
    const gchar *retval;

    debug ("_document_get_document_type\n");

    obj = g_object_get_data (G_OBJECT (document), PAPI_PYOBJECT);
    result = PyObject_CallMethod (obj, "document_get_document_type", NULL);
    if (!result)
        return NULL;

    retval = PyString_AsString (result);
    Py_DECREF (result);
    return retval;
}

/**
 * Wrapper for atk_document_get_document().
 */
static gpointer
_document_get_document (AtkDocument *document)
{
    PyObject *result;
    PyObject *obj;
    gpointer retval;

    debug ("_document_get_document\n");

    obj = g_object_get_data (G_OBJECT (document), PAPI_PYOBJECT);
    result = PyObject_CallMethod (obj, "document_get_document", NULL);
    if (!result)
        return NULL;

    retval = PyCObject_AsVoidPtr (result);
    Py_DECREF (result);
    return retval;
}

/**
 * Bind the different AtkDocumentIface interfaces to the wrapper
 * interfaces.
 */
static void
_atkdocumentiface_init (AtkDocumentIface *iface)
{
    iface->get_document_type = _document_get_document_type;
    iface->get_document = _document_get_document;
}

/**
 * Gets a modified interface for the AtkDocumentIface, which uses the
 * python wrappers as interfaces.
 */
const GInterfaceInfo*
atkdocumentiface_get_info (void)
{
    static const GInterfaceInfo info =
        {
            (GInterfaceInitFunc) _atkdocumentiface_init,
            (GInterfaceFinalizeFunc) NULL,
            NULL
        };
    return &info;
}

/**
 * Binds the AtkDocument methods to the passed object instance.
 */
void
atkdocumentiface_add_methods (PyObject *self)
{
    int i = 0;
    PyObject *func;
    
    while (_atkdocumentiface_methods[i].ml_name != NULL)
    {
        func = PyCFunction_New (&_atkdocumentiface_methods[i], NULL);
        PyObject_SetAttrString (self, _atkdocumentiface_methods[i].ml_name,
                                func);
        i++;
    }
}
